import { Entry } from "./entry.js";
import { HeadsCache } from "./heads-cache.js";
export class HeadsIndex {
    _id;
    _index = new Set();
    _gids;
    _headsCache;
    _blockstore;
    constructor(id) {
        this._gids = new Map();
        this._id = id;
    }
    async init(blockstore, cache, options = {}) {
        this._blockstore = blockstore;
        await this.reset(options?.entries || []);
        if (cache) {
            this._headsCache = new HeadsCache(this);
            return this._headsCache.init(cache);
        }
    }
    async load(options) {
        if (!this._headsCache || (this._headsCache.loaded && !options?.reload)) {
            return;
        }
        const heads = await this._headsCache?.load();
        if (!heads) {
            return;
        }
        const entries = await Promise.all(heads.map((x) => Entry.fromMultihash(this._blockstore, x, options)));
        await this.reset(entries);
        return entries;
    }
    get headsCache() {
        return this._headsCache;
    }
    close() {
        return this._headsCache?.close();
    }
    drop() {
        return this._headsCache?.drop();
    }
    get id() {
        return this._id;
    }
    get index() {
        return this._index;
    }
    get gids() {
        return this._gids;
    }
    get size() {
        return this._index.size;
    }
    async reset(entries, options = { cache: { reset: true, update: true } }) {
        this._index.clear();
        this._gids = new Map();
        if (entries) {
            await this.putAll(entries, options); // reset cache = true
        }
    }
    has(cid) {
        return this._index.has(cid);
    }
    async put(entry, options) {
        this._putOne(entry);
        if (options?.cache?.update) {
            await this._headsCache?.queue({ added: [entry] }, options.cache.reset);
        }
    }
    async putAll(entries, options) {
        this._putAll(entries);
        if (options?.cache?.update) {
            await this._headsCache?.queue({ added: entries }, options.cache.reset);
        }
    }
    async resetHeadsCache() {
        await this._headsCache?.queue({ added: [...this._index], removed: [] }, true);
    }
    async updateHeadsCache(change = {}, reset) {
        await this._headsCache?.queue(change, reset);
    }
    _putOne(entry) {
        if (!entry.hash) {
            throw new Error("Missing hash");
        }
        if (this._index.has(entry.hash)) {
            return;
        }
        this._index.add(entry.hash);
        if (!this._gids.has(entry.gid)) {
            this._gids.set(entry.gid, 1);
        }
        else {
            this._gids.set(entry.gid, this._gids.get(entry.gid) + 1);
        }
    }
    _putAll(entries) {
        for (const entry of entries) {
            this._putOne(entry);
        }
    }
    async del(entry, options) {
        const wasHead = this._index.delete(entry.hash);
        if (!wasHead) {
            return {
                lastWithGid: false,
                removed: false,
            };
        }
        const newValue = this._gids.get(entry.gid) - 1;
        const lastWithGid = newValue <= 0;
        if (newValue <= 0) {
            this._gids.delete(entry.gid);
        }
        else {
            this._gids.set(entry.gid, newValue);
        }
        if (!entry.hash) {
            throw new Error("Missing hash");
        }
        if (wasHead && options?.cache?.update) {
            await this._headsCache?.queue({ removed: [entry.hash] }, options.cache.reset);
        }
        return {
            removed: wasHead,
            lastWithGid: lastWithGid,
        };
        //     this._headsCache = undefined; // TODO do smarter things here, only remove the element needed (?)
    }
}
//# sourceMappingURL=heads.js.map