import { PublicKeyEncryptionResolver, SignatureWithKey } from "@dao-xyz/peerbit-crypto";
import type { BlockStore } from "@dao-xyz/libp2p-direct-block";
import LocalStore from "@dao-xyz/lazy-level";
import { EntryIndex } from "./entry-index.js";
import * as Sorting from "./log-sorting.js";
import { EncryptionTemplateMaybeEncrypted, Entry, Payload, CanAppend, EntryType } from "./entry.js";
import { HLC, LamportClock, Timestamp } from "./clock.js";
import { Encoding } from "./encoding.js";
import type { Identity } from "./identity.js";
import { CacheUpdateOptions, HeadsIndex } from "./heads.js";
import { Values } from "./values.js";
import { TrimOptions } from "./trim.js";
import { Change } from "./change.js";
import { EntryWithRefs } from "./entry-with-refs.js";
export type LogOptions<T> = LogProperties<T> & LogEvents<T>;
export type LogProperties<T> = {
    encryption?: PublicKeyEncryptionResolver;
    encoding?: Encoding<T>;
    clock?: LamportClock;
    sortFn?: Sorting.ISortFunction;
    trim?: TrimOptions;
    canAppend?: CanAppend<T>;
    cache?: (name: string) => Promise<LocalStore> | LocalStore;
};
export type LogEvents<T> = {
    onWrite?: (change: Entry<T>) => void;
    onChange?: (change: Change<T>) => void;
    onClose?: () => void;
};
export type AppendOptions<T> = {
    type?: EntryType;
    gidSeed?: Uint8Array;
    nexts?: Entry<any>[];
    identity?: Identity;
    signers?: ((data: Uint8Array) => Promise<SignatureWithKey>)[];
    reciever?: EncryptionTemplateMaybeEncrypted;
    onGidsShadowed?: (gids: string[]) => void;
    trim?: TrimOptions;
    timestamp?: Timestamp;
};
export declare class Log<T> {
    private _id;
    private _sortFn;
    private _storage;
    private _hlc;
    private _identity;
    private _entryIndex;
    private _headsIndex;
    private _values;
    private _nextsIndex;
    private _encryption?;
    private _encoding;
    private _trim;
    private _entryCache;
    private _canAppendInit?;
    private _canAppend?;
    private _onChangeInit?;
    private _onChange?;
    private _onWrite?;
    private _onClose?;
    private _joining;
    constructor(properties?: {
        id: Uint8Array;
    });
    init(store: BlockStore, identity: Identity, options?: LogOptions<T>): Promise<void>;
    setup(options?: {
        encoding?: Encoding<T>;
        canAppend?: CanAppend<T>;
        onChange?: (change: Change<T>) => void;
    }): Promise<void>;
    get initialized(): boolean;
    private _idString;
    get idString(): string;
    static createIdString(id: Uint8Array): string;
    get id(): Uint8Array;
    set id(id: Uint8Array);
    /**
     * Returns the length of the log.
     */
    get length(): number;
    get values(): Values<T>;
    /**
     * Checks if a entry is part of the log
     * @param {string} hash The hash of the entry
     * @returns {boolean}
     */
    has(cid: string): boolean;
    /**
     * Get all entries sorted. Don't use this method anywhere where performance matters
     */
    toArray(): Promise<Entry<T>[]>;
    /**
     * Returns the head index
     */
    get headsIndex(): HeadsIndex<T>;
    /**
     * Don't use this anywhere performance matters
     */
    getHeads(): Promise<Entry<T>[]>;
    /**
     * Returns an array of Entry objects that reference entries which
     * are not in the log currently.
     * @returns {Array<Entry<T>>}
     */
    getTails(): Promise<Entry<T>[]>;
    /**
     * Returns an array of hashes that are referenced by entries which
     * are not in the log currently.
     * @returns {Array<string>} Array of hashes
     */
    getTailHashes(): Promise<string[]>;
    /**
     * Get local HLC
     */
    get hlc(): HLC;
    get identity(): Identity;
    get storage(): BlockStore;
    get nextsIndex(): Map<string, Set<string>>;
    get entryIndex(): EntryIndex<T>;
    get encryption(): PublicKeyEncryptionResolver | undefined;
    get encoding(): Encoding<T>;
    get sortFn(): Sorting.ISortFunction;
    /**
     * Set the identity for the log
     * @param {Identity} [identity] The identity to be set
     */
    setIdentity(identity: Identity): void;
    /**
     * Find an entry.
     * @param {string} [hash] The hashes of the entry
     */
    get(hash: string, options?: {
        timeout?: number;
    }): Promise<Entry<T> | undefined>;
    traverse(rootEntries: Entry<T>[], amount?: number, endHash?: string): Promise<{
        [key: string]: Entry<T>;
    }>;
    getReferenceSamples(from: Entry<T>, options?: {
        pointerCount?: number;
        memoryLimit?: number;
    }): Promise<Entry<T>[]>;
    /**
     * Append an entry to the log.
     * @param {Entry} entry Entry to add
     * @return {Log} New Log containing the appended value
     */
    append(data: T, options?: AppendOptions<T>): Promise<{
        entry: Entry<T>;
        removed: Entry<T>[];
    }>;
    reset(entries: Entry<T>[], heads?: (string | Entry<T>)[]): Promise<void>;
    remove(entry: Entry<T> | Entry<T>[], options?: {
        recursively?: boolean;
    }): Promise<Change<T>>;
    iterator(options?: {
        from?: "tail" | "head";
        amount?: number;
    }): IterableIterator<string>;
    trim(option?: TrimOptions | undefined): Promise<Entry<T>[]>;
    /**
     *
     * @param entries
     * @returns change
     */
    join(entriesOrLog: (string | Entry<T> | EntryWithRefs<T>)[] | Log<T>, options?: {
        verifySignatures?: boolean;
        trim?: TrimOptions;
        timeout?: number;
    } & CacheUpdateOptions): Promise<void>;
    private joinEntry;
    private processEntry;
    deleteRecursively(from: Entry<any> | Entry<any>[], skipFirst?: boolean): Promise<Entry<T>[]>;
    delete(entry: Entry<any>): Promise<void>;
    /**
     * Returns the log entries as a formatted string.
     * @returns {string}
     * @example
     * two
     * └─one
     *   └─three
     */
    toString(payloadMapper?: (payload: Payload<T>) => string): Promise<string>;
    idle(): Promise<void>;
    close(): Promise<void>;
    drop(): Promise<void>;
    load(opts?: ({
        fetchEntryTimeout?: number;
    } & ({} | {
        heads?: true;
    })) & {
        reload: boolean;
    }): Promise<void>;
    static fromEntry<T>(store: BlockStore, identity: Identity, entryOrHash: string | string[] | Entry<T> | Entry<T>[], options?: {
        id?: Uint8Array;
        timeout?: number;
    } & LogOptions<T>): Promise<Log<T>>;
    /**
     * Find heads from a collection of entries.
     *
     * Finds entries that are the heads of this collection,
     * ie. entries that are not referenced by other entries.
     *
     * @param {Array<Entry<T>>} entries Entries to search heads from
     * @returns {Array<Entry<T>>}
     */
    static findHeads<T>(entries: Entry<T>[]): Entry<T>[];
    static findTails<T>(entries: Entry<T>[]): Entry<T>[];
    static findTailHashes(entries: Entry<any>[]): string[];
}
