import { Cache } from "@dao-xyz/cache";
import PQueue from "p-queue";
const trimOptionsEqual = (a, b) => {
    if (a.type === b.type) {
        if (a.type === "length" && b.type === "length") {
            return (a.from === b.from &&
                a.to === b.to &&
                a.filter?.canTrim === b.filter?.canTrim);
        }
        if (a.type === "bytelength" && b.type === "bytelength") {
            return (a.from === b.from &&
                a.to === b.to &&
                a.filter?.canTrim === b.filter?.canTrim);
        }
        if (a.type === "time" && b.type === "time") {
            return a.maxAge === b.maxAge && a.filter?.canTrim === b.filter?.canTrim;
        }
    }
    return false;
};
const trimOptionsStricter = (from, to) => {
    if (from.type !== to.type || from.filter?.canTrim !== to.filter?.canTrim) {
        // TODO also check ttl?
        return true; // we don't really know
    }
    if ((from.type === "bytelength" || from.type === "length") &&
        (to.type === "bytelength" || to.type === "length")) {
        if (from.to > to.to) {
            return true;
        }
        const fromFrom = from.from || from.to;
        const fromTo = to.from || to.to;
        return fromFrom > fromTo;
    }
    else {
        if (from.type === "time" && to.type === "time") {
            return from.maxAge > to.maxAge;
        }
    }
    throw new Error("Unexpected");
};
export class Trim {
    _trim;
    _canTrimCacheLastNode;
    _trimLastHead;
    _trimLastTail;
    _trimLastOptions;
    _trimLastSeed;
    _canTrimCacheHashBreakpoint;
    _log;
    _queue;
    constructor(log, options) {
        this._log = log;
        this._trim = options;
        this._canTrimCacheHashBreakpoint = new Cache({ max: 1e5 });
        this._queue = new PQueue({ concurrency: 1 });
    }
    deleteFromCache(entry) {
        if (this._canTrimCacheLastNode?.value.hash === entry.hash) {
            this._canTrimCacheLastNode = this._canTrimCacheLastNode.prev;
        }
    }
    get options() {
        return this._trim;
    }
    async trimTask(option = this._trim) {
        if (!option) {
            return [];
        }
        ///  TODO Make this method less ugly
        const deleted = [];
        let done;
        const values = this._log.values();
        if (option.type === "length") {
            const to = option.to;
            const from = option.from ?? to;
            if (values.length < from) {
                return [];
            }
            done = () => values.length <= to;
        }
        else if (option.type == "bytelength") {
            // prune to max sum payload sizes in bytes
            const byteLengthFrom = option.from ?? option.to;
            if (values.byteLength < byteLengthFrom) {
                return [];
            }
            done = () => values.byteLength <= option.to;
        }
        else if (option.type == "time") {
            const s0 = BigInt(+new Date() * 1e6);
            const maxAge = option.maxAge * 1e6;
            done = async () => {
                if (!values.tail) {
                    return true;
                }
                const nodeValue = await values.getEntry(values.tail);
                if (!nodeValue) {
                    return true;
                }
                return s0 - nodeValue.metadata.clock.timestamp.wallTime < maxAge;
            };
        }
        else {
            return [];
        }
        const tail = values.tail;
        if (this._trimLastOptions &&
            trimOptionsStricter(this._trimLastOptions, option)) {
            this._canTrimCacheHashBreakpoint.clear();
        }
        const seed = option.filter?.cacheId?.();
        const cacheProgress = seed != null;
        let changed = false;
        if (seed !== this._trimLastSeed || !cacheProgress) {
            // Reset caches
            this._canTrimCacheHashBreakpoint.clear();
            this._canTrimCacheLastNode = undefined;
            changed = true;
        }
        else {
            const trimOptionsChanged = !this._trimLastOptions ||
                !trimOptionsEqual(this._trimLastOptions, option);
            const changed = this._trimLastHead !== values.head ||
                this._trimLastTail !== values.tail ||
                trimOptionsChanged;
            if (!changed) {
                return [];
            }
        }
        let node = this._canTrimCacheLastNode || tail; // TODO should we do this._canTrimCacheLastNode?.prev instead ?
        let lastNode = node;
        let looped = false;
        const startNode = node;
        let canTrimByGid = undefined;
        // TODO only go through heads?
        //console.log("START TRIM", await done(), option.filter?.canTrim)
        while (node &&
            !(await done()) &&
            values.length > 0 &&
            node &&
            (!looped || node !== startNode)) {
            let deleteAble = true;
            if (option.filter?.canTrim) {
                canTrimByGid = canTrimByGid || new Map();
                deleteAble = canTrimByGid.get(node.value.gid);
                if (deleteAble === undefined) {
                    deleteAble = await option.filter?.canTrim(node.value.gid);
                    canTrimByGid.set(node.value.gid, deleteAble);
                }
                if (!deleteAble && cacheProgress) {
                    // ignore it
                    this._canTrimCacheHashBreakpoint.add(node.value.hash, true);
                }
            }
            // Delete, and update current node
            if (deleteAble) {
                // Do this before deleteNode, else prev/next might be gone!
                const prev = node.prev;
                const next = node.next;
                const entry = await this._log.deleteNode(node);
                if (entry) {
                    deleted.push(entry);
                }
                node = prev;
                // If we don't do this, we might, next time start to iterate from a node that does not exist
                // we do prev 'or' next because next time we want to start as close as possible to where we left of
                lastNode = prev || next;
            }
            else {
                lastNode = node;
                node = node?.prev;
            }
            if (!node) {
                if (!looped && changed && !cacheProgress) {
                    node = tail;
                    looped = true;
                }
                else {
                    break;
                }
            }
        }
        // remember the node where we started last time from
        this._canTrimCacheLastNode = node || lastNode;
        this._trimLastHead = values.head;
        this._trimLastTail = values.tail;
        this._trimLastOptions = option;
        this._trimLastSeed = seed;
        return deleted;
    }
    /**
     * @param options
     * @returns deleted entries
     */
    async trim(option = this._trim) {
        const result = await this._queue.add(() => this.trimTask(option));
        if (result instanceof Object) {
            return result;
        }
        throw new Error("Something when wrong when trimming");
    }
}
//# sourceMappingURL=trim.js.map